##############################################################################
#
# PhotoIptcExif ProductExtension
#
# IPTC and EXIF reading support.
# A product extension to Ron Bickers' Photo Product 1.2.3, for Zope
#
# Copyright (c) Augusto Cabrera   All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Augusto Cabrera    ausum_studio@hotmail.com
#    February 23rd, 2002
#
###########################################################################
#
# This hotfix contains modifications to methods from the Photo Product,
# and code from the following authors: 
# 
# Ron Bickers' Photo Product:
# http://www.zope.org/Members/rbickers/Photo
# 
# Fredrik Lundh   fredrik@pythonware.com, for the IPTC reader code
# within MetadataHandler.py
#
# Robert F. Tobler, for the EXIF reader code (exif.py and exiftags.py).
#
###########################################################################


import sys, string
from OFS.Image import Image, cookId, manage_addImage
from Acquisition import Implicit
from Globals import Persistent
from ImageFile import ImageFile
from OFS.SimpleItem import Item
from OFS.PropertyManager import PropertyManager
from OFS.PropertySheets import PropertySheet
from AccessControl.Role import RoleManager
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, DTMLFile
from Products.PythonScripts.standard import html_quote
try:
    from webdav.WriteLockInterface import WriteLockInterface
    from webdav.Lockable import ResourceLockedError
except ImportError: pass
from cStringIO import StringIO
try: 
    from PIL.Image import *
    import PIL.Image
except ImportError: pass
import time
from exif import parse
from MetadataHandler import iptcexifHandler, li, get_info
from Products import Photo

defaultdisplays = {'thumbnail': (128,128),
                   'xsmall': (200,200),
                   'small': (320,320),
                   'medium': (480,480),
                   'large': (768,768),
                   'xlarge': (1024,1024)
                  }

def __init__(self, id, title, file, content_type='', precondition='',
             store='Image', engine='ImageMagick', quality=75, pregen=0,
             timeout=0, iptcproperties=0, printdisplaycreation=0 ):

    self.__version__ = '1.2.3'
    self.id = id
    self.title = title   

    # Sheet to store photo settings.
    self.propertysheets.manage_addPropertySheet('photoconf', 'photoconf')
    photoconf = self.propertysheets.get('photoconf')
    photoconf.manage_addProperty('store', store, 'string')
    photoconf.manage_addProperty('engine', engine, 'string')
    photoconf.manage_addProperty('quality', quality, 'int')
    photoconf.manage_addProperty('pregen', pregen, 'boolean')
    photoconf.manage_addProperty('timeout', timeout, 'int')
    photoconf.manage_addProperty('iptcproperties', iptcproperties, 'boolean')
    photoconf.manage_addProperty('printdisplaycreation', printdisplaycreation, 'boolean')

    # Initialize with default hardcoded sizes.
    self._displays = defaultdisplays.copy()
    self._photos = {}
    
minus = ImageFile('www/mi.gif',globals())
plus = ImageFile('www/pl.gif',globals())
non = ImageFile('www/non.gif',globals())
errorgif = ImageFile('www/error.gif',globals())

def width(self):
    """Original photo width within Photo."""
    return self._original.nwidth
    # return self._original.width

def height(self):
    """Original photo height within Photo."""
    return self._original.nheight
    # return self._original.height

def anIPTCproperty(self, id):
    """Returns true if id is an IPTC field name."""
    return id in [v for k, v in li]

def exifHandler(self, option=0, id=None, ed=None):
    """Handle EXIF data from Tiffs and JPEGs."""        

    li = [('PhotoName','Photo Name'),('Make', 'Camera Maker'),('Model','Camera Model'),
          ('DateTime','Date & Time'),('ExposureTime','Exposure Time'), ('FNumber','Aperture Value'),
          ('ExposureProgram','Exposure Program'), ('ExposureBiasValue', 'Exposure Bias'),
          ('MeteringMode','Metering Mode'),('Flash','Flash'),('FocalLength','Focal Length')]
    if option == 0:
        return li
    elif option == 1:
        a_basic_ex = id in [k for k, v in li]
        return a_basic_ex
    elif option == 2 and ed is not None:
        ed = ed        
        nli = []
        for k,v in li:
            try:
                rdic = str(ed[k])
                del ed[k]
            except:
                rdic = 'N/A'
            nli.append((v,rdic))
        for k,v in ed.items():
            if k[:3]=='MN_':
                nli.append((k,v))
                del ed[k]
        nli2 = ed.items()        
        return nli + nli2

def _resize(self, display, width, height, engine='ImageMagick', quality=75):
    """Resize and resample photo."""
    origimg = self._original
    id = self.id
    newimg = StringIO()
    if engine == 'PIL':  # Use PIL
        img = PIL.Image.open(origimg._PILdata())
        # fmt = img.format    # This line disabled because we'll only use the jpeg format.
        try:
            if img.mode is not 'RGB':
                # We convert other modes to RGB before resizing and saving.
                img = img.convert('RGB').resize((width, height), resample=BICUBIC)
                img.save(newimg, format='jpeg', quality=quality)
            else:
                img = img.resize((width, height), resample=BICUBIC)
                img.save(newimg, format='jpeg', quality=quality)
            if verbosedisplays: print "    %s image: ( %s x %s ).......... created" % (display, width, height)
        except:
            if verbosedisplays: print "    %s image: ( %s x %s ).......... not created" % (display, width, height)
    elif engine == 'ImageMagick':  # Use ImageMagick            
        if sys.platform == 'win32':
            from win32pipe import popen2            
            imgin, imgout = popen2('convert -quality %s -geometry %sx%s! - jpg:-'  # Jpeg displays
                                   % (quality, width, height), 'b')
        else:
            from popen2 import popen2
            imgout, imgin = popen2('convert -quality %s -geometry %sx%s! - jpg:-'
                                   % (quality, width, height))        
        imgin.write(origimg._IMdata())
        imgin.close()
        imgread = imgout.read()
        newimg.write(imgread)
        if verbosedisplays:
            if imgread != '':
                print "    %s image: ( %s x %s ).......... created" % (display, width, height)
            else:
                print "    %s image: ( %s x %s ).......... not created" % (display, width, height)
        imgout.close()
    newimg.seek(0)
    return newimg            

def PUT(self, REQUEST, RESPONSE):
    """Handle HTTP PUT requests."""
    file = REQUEST['BODYFILE']
    im = PIL.Image.open(file)    
    self.dav__init(REQUEST, RESPONSE)
    iptcproperties = self.propertysheets.get('photoconf').getProperty('iptcproperties')    
    if hasattr(self, 'dav__simpleifhandler'):
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
    get_info(im)
    file.seek(0)
    global verbosedisplays
    if self.propertysheets.get('photoconf').getProperty('printdisplaycreation'):        
        verbosedisplays = 1
    else: verbosedisplays = 0
    if hasattr(self, '_original'):
        # Updating existing Photo
        readfile = file.read()                        
        if iptcproperties:
            iptcexifHandler(self, im, cp=1 )
        else:
            iptcexifHandler(self, im, cp=0, justexif=1 )        
        self._original.manage_upload(file, self.content_type())
        if self._validImage():
            self._makeDisplayPhotos()
    else:
        
        # Adding a new Photo.
        # At this point, the object is not yet in its final context.
        # Since ExtImage needs to know the URL of the new Photo object
        # so it can determine where to create new files, we store
        # the image data and let manage_afterAdd() generate the displays
        # once the Photo is in its final context.
        self._data = readfile = file.read()
        global imgsizes
        imgsizes = im.width, im.height
        # IPTC and EXIF data handling. 
        if iptcproperties:
            iptcexifHandler(self, im, cp=1 )
        else:
            iptcexifHandler(self, im, cp=0, justexif=1 )
    RESPONSE.setStatus(204)
    return RESPONSE

def manage_editPhoto(self, file='', oid=0, REQUEST=None):
    """Changes Photo information."""
    iptcproperties = self.propertysheets.get('photoconf').getProperty('iptcproperties')
    if hasattr(self, 'wl_isLocked') and self.wl_isLocked():
        raise ResourceLockedError, "Photo is locked via WebDAV."    
    self.manage_changeProperties(REQUEST)
    global verbosedisplays
    if file and file.filename:
        file.seek(0)
        try:
            im = PIL.Image.open(file)
        except IOError:
            raise
        get_info(im)
        global imgsizes
        imgsizes = im.width, im.height
        file.seek(0)        
        if iptcproperties and oid:   #Using iptcexifHandler, wich uses PIL
            readfile = self._data = file.read()
            store = self.propertysheets.get('photoconf').getProperty('store')
            if store == 'Image': from Products.Photo.PhotoImage import PhotoImage
            elif store == 'ExtImage': from Products.Photo.ExtPhotoImage import PhotoImage
            self._original = PhotoImage(self.id, self.title, path=self.absolute_url(1))
            self._original.manage_upload(StringIO(self._data), self.content_type())
            delattr(self, '_data')
            iptcexifHandler(self, im, cp=1 )                
        else:           
            readfile = file.read()
            iptcexifHandler(self, im, cp=0, justexif=1 )
            self._original.manage_upload(file, self.content_type())
        self._original.nwidth, self._original.nheight = imgsizes
        if self._validImage():
            self._makeDisplayPhotos()
        else:
            try: self._makeDisplayPhotos() # FIXME. This is to force other formats to parse.
            except: pass
    if REQUEST is not None:            
        # Forcing the page to refresh. Helpful within IE browsers when caching is disabled.        
        posturl ='manage_editPhotoForm?manage_tabs_message=Photo+information+updated.'
        return self.REQUEST['RESPONSE'].redirect('%s/%s' % (self.absolute_url(), posturl))


def manage_afterAdd(self, item, container):
    """Handle pasting of new photos."""
    global verbosedisplays
    
    # Init properties and displays from Photo Folder if present
    parent = self.aq_parent
    if parent.meta_type == 'Photo Folder':
        props = parent.propertysheets.get('photos')
        for propid, value in props.propertyItems():
            try: self.manage_addProperty(propid, value, props.getPropertyType(propid))
            except: pass
        self._displays = parent._displays.copy()
        
    if not hasattr(self, '_original'):
        # Added Photo (vs. imported)
        # See note in PUT()     
        store = self.propertysheets.get('photoconf').getProperty('store')
        if store == 'Image':
            from Products.Photo.PhotoImage import PhotoImage
        elif store == 'ExtImage':
            from Products.Photo.ExtPhotoImage import PhotoImage
        self._original = PhotoImage(self.id, self.title, path=self.absolute_url(1))
        self._original.manage_upload(StringIO(self._data), self.content_type())      
        delattr(self, '_data')
        self._original.nwidth, self._original.nheight = imgsizes
        if verbosedisplays: print self.id
        if self._validImage():
            self._makeDisplayPhotos()
        else:
            try: self._makeDisplayPhotos() # FIXME. This is to force other formats to parse.
            except: pass
    self._original.manage_afterAdd(item, container)
    self._original.nwidth, self._original.nheight = imgsizes 
    if hasattr(self, '_photos'):
        for photo in self._photos.values():
            photo.manage_afterAdd(item, container)
    if verbosedisplays: verbosedisplays = None    


manage_editPhotoForm = DTMLFile('dtml/editPhotoForm', globals())
manage_editSettingsForm = DTMLFile('dtml/editSettingsForm', globals())
manage_addPhotoForm = DTMLFile('dtml/addPhotoForm', globals())
manage_propertiesForm = DTMLFile('dtml/propertiesForm', globals())


def manage_addPhoto(dispatcher, id, title, file,
                    content_type='', precondition='',
                    store='Image', engine='ImageMagick', quality=75,
                    timeout=0, pregen=0, iptcproperties=0, REQUEST=None):
    """Add Photo object."""
        
    id = cookId(id, title, file)[0]
    dest = dispatcher.Destination()
    photo = Photo.Photo(id, title, '', content_type, precondition,
                  store, engine, quality, pregen, timeout, iptcproperties)
    global verbosedisplays
    verbosedisplays = 0
    import types
    if type(file) == types.InstanceType:
        photo._data = file.read()
    else:
        photo._data = str(file)
        file = _to_file(photo._data)
    file.seek(0)
    try:
        img = PIL.Image.open(file)
    except IOError:
        raise

    file.seek(0)
    get_info(img)
    global imgsizes
    imgsizes = img.width, img.height
    file.seek(0)
    
    dest._setObject(id, photo) 
    self = dest._getOb(id)
   
    if iptcproperties:
        iptcexifHandler(self, img, cp=1 )  
    else:          
        iptcexifHandler(self, img, cp=0, justexif=1 )        
    if REQUEST is not None:
        try:    url=dispatcher.DestinationURL()
        except: url=REQUEST['URL1']
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
    return id



Photo.Photo.__init__ = __init__
Photo.Photo.minus = minus
Photo.Photo.plus = plus
Photo.Photo.non = non

Photo.Photo.anIPTCproperty = anIPTCproperty
Photo.Photo.exifHandler = exifHandler
Photo.Photo._resize = _resize

Photo.Photo.width = width
Photo.Photo.height = height

Photo.Photo.PUT = PUT
Photo.Photo.manage_editPhoto = manage_editPhoto
Photo.Photo.manage_afterAdd = manage_afterAdd

Photo.Photo.manage_editPhotoForm = manage_editPhotoForm
Photo.Photo.manage_propertiesForm = manage_propertiesForm
Photo.Photo.manage_addPhotoForm = manage_addPhotoForm
Photo.Photo.manage_editSettingsForm = manage_editSettingsForm

Photo.manage_addPhotoForm = manage_addPhotoForm
Photo.manage_addPhoto = manage_addPhoto


def init(self, id, title, store='Image', engine='ImageMagick', quality=75,
             pregen=0, timeout=0, iptcproperties=0, printdisplaycreation=0):

    self.__version__ = '1.2.3'
    self.id = id
    self.title = title
    self.image = ''

    # Sheet to store containing photo default settings.
    self.propertysheets.manage_addPropertySheet('photoconf', 'photoconf')
    photoconf = self.propertysheets.get('photoconf')
    photoconf.manage_addProperty('store', store, 'string')
    photoconf.manage_addProperty('engine', engine, 'string')
    photoconf.manage_addProperty('quality', quality, 'int')
    photoconf.manage_addProperty('pregen', pregen, 'boolean')
    photoconf.manage_addProperty('timeout', timeout, 'int')
    photoconf.manage_addProperty('iptcproperties', iptcproperties, 'boolean')
    photoconf.manage_addProperty('printdisplaycreation', printdisplaycreation, 'boolean')

    # Sheet to store containing photo properties.
    self.propertysheets.manage_addPropertySheet('photos', 'photos')

    # Initialize with default hardcoded sizes.
    self._displays = defaultdisplays.copy()

manage_editPhotoPropertiesForm = DTMLFile('dtml/editPhotoPropertiesForm', globals())
manage_addPhotoFolderForm = DTMLFile('dtml/addPhotoFolderForm', globals())

from Products.Photo.PhotoFolder import PhotoFolder

def manage_addPhotoFolder(dispatcher, id, title, store='Image',
                          engine='ImageMagick', quality=75, pregen=0, timeout=0,
                          createsamples=0, iptcproperties=0, printdisplaycreation=0, REQUEST=None):
    """Add Photo Folder object"""
    dest = dispatcher.Destination()
    dest._setObject(id, PhotoFolder(id, title, store, engine, quality, pregen,
                                    timeout, iptcproperties, printdisplaycreation))
    if createsamples:
        self = dest._getOb(id)
        sampleView = DTMLFile('dtml/sampleView', globals())
        self.manage_addDTMLMethod('view', '')
        self._getOb('view').manage_edit(sampleView, 'Sample View')
        sampleFolderView = DTMLFile('dtml/sampleFolderView', globals())
        self.manage_addDTMLMethod('index_html', '')
        self._getOb('index_html').manage_edit(sampleFolderView, 'Sample Folder View')

    if REQUEST is not None:
        try:    url=dispatcher.DestinationURL()
        except: url=REQUEST['URL1']
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
    return id

unsupported = ['image/pict',]

def PUT_factory(self, name, typ, body):
    """Create Photo objects by default for image types."""
    if typ in ['image/jpeg', 'image/png', 'image/gif']:
        
        store = self.propertysheets.get('photoconf').getProperty('store')
        photo = Photo.Photo(name, '', '', store=store)
        # Init properties with defaults
        props = self.propertysheets.get('photos')
        for propid, value in props.propertyItems():
            try: photo.manage_addProperty(propid, value, props.getPropertyType(propid))
            except: pass
        # Init settings with defaults
        photoconf = self.propertysheets.get('photoconf')
        settings = {}
        for propid, value in photoconf.propertyItems():
            settings[propid] = value
        photo.propertysheets.get('photoconf').manage_changeProperties(settings)
        # Init displays with defaults
        photo._displays = self._displays.copy()
        return photo
    
    elif typ not in unsupported:
        file = _to_file(body)  # Is there a better way to perform this?   
        try:
            PIL.Image.open(file) # We want ALL the files to parse. Not image files won't create
            file.seek(0)         # a Photo object, but they will do file objects.

            store = self.propertysheets.get('photoconf').getProperty('store')
            photo = Photo.Photo(name, '', '', store=store)
            # Init properties with defaults
            props = self.propertysheets.get('photos')
            for propid, value in props.propertyItems():
                try: photo.manage_addProperty(propid, value, props.getPropertyType(propid))
                except: pass
            # Init settings with defaults
            photoconf = self.propertysheets.get('photoconf')
            settings = {}
            for propid, value in photoconf.propertyItems():
                settings[propid] = value
            photo.propertysheets.get('photoconf').manage_changeProperties(settings)
            # Init displays with defaults
            photo._displays = self._displays.copy()
            return photo      
        except: pass
    return None

def _to_file(readfile):      # FIXME, if there's a better way to return a file, given its
    import __builtin__       # already read content.
    em = __builtin__.open('temp_filename', 'wb')
    em.write(readfile)
    em.close()
    file = __builtin__.open('temp_filename', 'rb')
    file.seek(0)
    return file


Photo.PhotoFolder.__init__ = init
Photo.PhotoFolder.manage_editPhotoPropertiesForm = manage_editPhotoPropertiesForm
Photo.PhotoFolder.PUT_factory = PUT_factory
Photo.manage_addPhotoFolderForm = manage_addPhotoFolderForm
Photo.manage_addPhotoFolder = manage_addPhotoFolder
